/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { AttachState } from "@fluidframework/container-definitions";
import { createDetachedContainer, loadExistingContainer, } from "@fluidframework/container-loader/internal";
import { createDOProviderContainerRuntimeFactory, createFluidContainer, createServiceAudience, } from "@fluidframework/fluid-static/internal";
import { RouterliciousDocumentServiceFactory } from "@fluidframework/routerlicious-driver/internal";
import { wrapConfigProviderWithDefaults } from "@fluidframework/telemetry-utils/internal";
import { InsecureTinyliciousTokenProvider, InsecureTinyliciousUrlResolver, createTinyliciousCreateNewRequest, } from "@fluidframework/tinylicious-driver/internal";
import { createTinyliciousAudienceMember } from "./TinyliciousAudience.js";
/**
 * Provides the ability to have a Fluid object backed by a Tinylicious service.
 *
 * @see {@link https://fluidframework.com/docs/testing/tinylicious/}
 *
 * @sealed
 * @public
 */
export class TinyliciousClient {
    /**
     * Creates a new client instance using configuration parameters.
     * @param properties - Optional. Properties for initializing a new TinyliciousClient instance
     */
    constructor(properties) {
        this.logger = properties?.logger;
        const tokenProvider = new InsecureTinyliciousTokenProvider();
        this.urlResolver = new InsecureTinyliciousUrlResolver(properties?.connection?.port, properties?.connection?.domain);
        this.documentServiceFactory = new RouterliciousDocumentServiceFactory(properties?.connection?.tokenProvider ?? tokenProvider);
    }
    /**
     * Creates a new detached container instance in Tinylicious server.
     * @param containerSchema - Container schema for the new container.
     * @param compatibilityMode - Compatibility mode the container should run in.
     * @returns New detached container instance along with associated services.
     */
    async createContainer(containerSchema, compatibilityMode) {
        const loaderProps = this.getLoaderProps(containerSchema, compatibilityMode);
        // We're not actually using the code proposal (our code loader always loads the same module
        // regardless of the proposal), but the Container will only give us a NullRuntime if there's
        // no proposal.  So we'll use a fake proposal.
        const container = await createDetachedContainer({
            ...loaderProps,
            codeDetails: {
                package: "no-dynamic-package",
                config: {},
            },
        });
        /**
         * See {@link FluidContainer.attach}
         */
        const attach = async () => {
            if (container.attachState !== AttachState.Detached) {
                throw new Error("Cannot attach container. Container is not in detached state.");
            }
            const request = createTinyliciousCreateNewRequest();
            await container.attach(request);
            if (container.resolvedUrl === undefined) {
                throw new Error("Resolved Url not available on attached container");
            }
            return container.resolvedUrl.id;
        };
        const fluidContainer = await createFluidContainer({
            container,
        });
        fluidContainer.attach = attach;
        const services = this.getContainerServices(container);
        return { container: fluidContainer, services };
    }
    /**
     * Accesses the existing container given its unique ID in the tinylicious server.
     * @param id - Unique ID of the container.
     * @param containerSchema - Container schema used to access data objects in the container.
     * @param compatibilityMode - Compatibility mode the container should run in.
     * @returns Existing container instance along with associated services.
     */
    async getContainer(id, containerSchema, compatibilityMode) {
        const loaderProps = this.getLoaderProps(containerSchema, compatibilityMode);
        const container = await loadExistingContainer({ ...loaderProps, request: { url: id } });
        const fluidContainer = await createFluidContainer({
            container,
        });
        const services = this.getContainerServices(container);
        return { container: fluidContainer, services };
    }
    // #region private
    getContainerServices(container) {
        return {
            audience: createServiceAudience({
                container,
                createServiceMember: createTinyliciousAudienceMember,
            }),
        };
    }
    getLoaderProps(schema, compatibilityMode) {
        const containerRuntimeFactory = createDOProviderContainerRuntimeFactory({
            schema,
            compatibilityMode,
        });
        const load = async () => {
            return {
                module: { fluidExport: containerRuntimeFactory },
                details: { package: "no-dynamic-package", config: {} },
            };
        };
        const codeLoader = { load };
        const client = {
            details: {
                capabilities: { interactive: true },
            },
            permission: [],
            scopes: [],
            user: { id: "" },
            mode: "write",
        };
        const featureGates = {
            // T9s client requires a write connection by default
            "Fluid.Container.ForceWriteConnection": true,
        };
        const loaderProps = {
            urlResolver: this.urlResolver,
            documentServiceFactory: this.documentServiceFactory,
            codeLoader,
            logger: this.logger,
            options: { client },
            configProvider: wrapConfigProviderWithDefaults(/* original */ undefined, featureGates),
        };
        return loaderProps;
    }
}
//# sourceMappingURL=TinyliciousClient.js.map